// -*- Mode:C++ -*-
//Header:
//File: AuElman7L.h
//Title: Autistic Elman Net with 5 layers
//Author: NODA, Itsuki
//Date: 2000/01/05
//

//ModifyHistory:
// 2000/01/05: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_AuElman7L_h_
#define _Nnn_AuElman7L_h_
////////////////////////////////////////////////////////////////////////

#include "BpNetBase.h"

namespace Nnn {

   using namespace Itk ;

   //======================================================================
   // Autistic Elman 7 Layer (5layer + 2 delayed context)

   class AuElman7L : public BpNetBase {

      public:
	 //--------------------------------------------------
	 // constants

	 //------------------------------
	 // amount of adjust context

	 static const Val adjustRate = 1.0;

	 //------------------------------
	 // for save/load/scan

	 static char* const _sformTermHead = "AuElman7L" ; 
	 virtual char* sformTermHead() const {
	    return _sformTermHead ;
	 } ;

	 static char * const nameContext0 = "context0" ;
	 static char * const nameInter0 = "inter0" ;
	 static char * const nameContext = "context" ;
	 static char * const nameInter1 = "inter1" ;
	 static char * const nameHidden = "hidden" ;
	 static char * const nameInter2 = "inter2" ;
	 static char * const nameOutput = "output" ;


	 //--------------------------------------------------
	 // member
      public:
	 BpLayer * context0 ;
	 BpLayer * inter0 ;
	 BpLayer * context ;
	 BpLayer * inter1 ;
	 BpLayer * hidden ;
	 BpLayer * inter2 ;
	 BpLayer * output ;

	 BpLLink * con0_int0 ;
	 BpLLink * int0_con ;
	 BpLLink * con_int1 ;
	 BpLLink * int1_hid ;
	 BpLLink * hid_int2 ;
	 BpLLink * int2_out ;

	 ValVector * ctxvec ;
	 ValVector * intvec ;
	 ValVector * ctx0vec ;

	 //--------------------------------------------------
	 // constructor
      public:
	 AuElman7L() : ctxvec(), intvec(), ctx0vec() { init() ; } ;

	 //--------------------------------------------------
	 // alloc new memory
     public:
	 void alloc(const SString & nm,
		    const UInt & int1sz,
		    const UInt & hidsz,
		    const UInt & int2sz,
		    const UInt & outsz,
		    const Val & tempval = NNN_DFLT_TEMP,
		    const Val & epsval = NNN_DFLT_EPS,
		    const Val & alphaval = NNN_DFLT_ALPHA) {
	    init(nm,tempval,epsval,alphaval) ;
	    assignLayerLLink(int1sz,hidsz,int2sz,outsz) ;
	 } ;
	 
	 //--------------------------------------------------
	 // assignLayer (l-s-l-s-l)
      public:
	 void assignLayerLLink(const UInt & int1sz,
			       const UInt & hidsz, 
			       const UInt & int2sz,
			       const UInt & outsz) {
	    context0 = newLayer(nameContext0,hidsz,BpLayer::FT_linear) ;
	    inter0 = newLayer(nameInter0,int1sz,BpLayer::FT_sigmoid) ;
	    context = newLayer(nameContext,hidsz,BpLayer::FT_linear) ;
	    inter1 = newLayer(nameInter1,int1sz,BpLayer::FT_sigmoid) ;
	    hidden = newLayer(nameHidden,hidsz,BpLayer::FT_linear) ;
	    inter2 = newLayer(nameInter2,int2sz,BpLayer::FT_sigmoid) ;
	    output = newLayer(nameOutput,outsz,BpLayer::FT_linear) ;
	    
	    con0_int0 = newLLink(context0,inter0) ;
	    int0_con = newLLink(inter0,context);
	    con_int1 = newLLink(context,inter1) ;
	    int1_hid = newLLink(inter1,hidden);
	    hid_int2 = newLLink(hidden,inter2) ;
	    int2_out = newLLink(inter2,output) ;
	 } ;

	 //--------------------------------------------------
	 // fix buffer vec
      public:
	 Bool fixBufferVec() {
	    if(isNull(ctxvec)) {
	       ctxvec = new ValVector(hidden->n) ;
	       ctx0vec = new ValVector(hidden->n) ;
	       intvec = new ValVector(inter1->n) ;
	       return True ;
	    } else {
	       return False ;
	    }
	 } ;
	    

	 //--------------------------------------------------
	 // propFore
      public:
	 void propFore(Bool firstp = True) {
	    if(firstp) {
	       fixBufferVec() ;

	       hidden->getOut(ctxvec) ;
	       context->setOut(ctxvec) ;
	    }
	    
	    inter1->propFore(temp) ;
	    hidden->propFore(temp) ;
	    inter2->propFore(temp) ;
	    output->propFore(temp) ;
	 } ;
	    
	 //--------------------------------------------------
	 // propBack
      public:
	 void propBack(const ValVector & tchvec) {
	    output->propBack(tchvec,temp) ;
	    inter2->propBack(temp) ;
	    hidden->propBack(temp) ;
	    inter1->propBack(temp) ;
	    context->propBack(temp) ;
	    inter0->propBack(temp) ;
	 } ;

	 //--------------------------------------------------
	 // adjust Context
      public:
	 void adjustContext(const Val & eps) {
	    context->adjustOutputByError(eps) ;
	 } ;

	 //--------------------------------------------------
	 // learn
      public:
	 Flt learn() {
	    Val r = 0.0 ;
	    r += inter1->learn(eps,alpha) ;
	    r += hidden->learn(eps,alpha) ;
	    r += inter2->learn(eps,alpha) ;
	    r += output->learn(eps,alpha) ;

	    r += con_int1->learn(eps,alpha) ;
	    r += int1_hid->learn(eps,alpha) ;
	    r += hid_int2->learn(eps,alpha) ;
	    r += int2_out->learn(eps,alpha) ;

	    r += con0_int0->learn(eps,alpha) ;
	    r += int0_con->learn(eps,alpha) ;

	    return r ;
	 } ;

	 //--------------------------------------------------
	 // run
      public:
	 void run() { return propFore(True) ; } ;
	 void run(const ValVector & tchvec) { 
	    propFore(True) ;
	    propBack(tchvec) ;
	    adjustContext(adjustRate) ;
	    propFore(False) ;
	 } ;
	 void runLearn(const ValVector & tchvec) {
	    fixBufferVec() ;
	    context0->setOut(ctx0vec) ;
	    inter0->setOut(intvec) ;

	    propFore(True) ;
	    context->getOut(ctx0vec) ;
	    inter1->getOut(intvec) ;

	    propBack(tchvec) ;
	    learn() ;
	    adjustContext(adjustRate) ;

	    propFore(False) ;
	    propBack(tchvec) ;
	    learn() ;
	 } ;

	 //--------------------------------------------------
	 // get sum of square error
      public:
	 Flt sumError() const { return output->sumError() ; } ;

	 //--------------------------------------------------
	 // get output
      public:
	 void getNetOutput(ValVector & outvec) const { 
	    output->getNetOutput(outvec) ;
	 } ;

	 //--------------------------------------------------
	 // scan
      public:
	 virtual void scan(StrForm & sform) {
	    BpNetBase::scan(sform) ;

	    context0 = layer(nameContext0) ;
	    inter0 = layer(nameInter0) ;
	    context = layer(nameContext) ;
	    inter1 = layer(nameInter1) ;
	    hidden = layer(nameHidden) ;
	    inter2 = layer(nameInter2) ;
	    output = layer(nameOutput) ;

	    con0_int0 = link(context0,inter0) ;
	    int0_con = link(inter0,context) ;
	    con_int1 = link(context,inter1) ;
	    int1_hid = link(inter1,hidden) ;
	    hid_int2 = link(hidden,inter2) ;
	    int2_out = link(inter2,output) ;
	 } ;
   } ;

} ;
////////////////////////////////////////////////////////////////////////
#endif


