// -*- Mode:C++ -*-
//Header:
//File: BpLayer.cc
//Author: NODA, Itsuki
//Date: 1999/12/31
//

//ModifyHistory:
// 1999/12/31: Start to create this file
//EndModifyHistory:

#include "BpLayer.h"

namespace Nnn {

   //======================================================================
   // Execution

   //----------------------------------------------------------------------
   // forward processing

   //--------------------------------------------------
   // set net input (set Input Vector to input layer)

   void BpLayer::setNetInput(const ValVector & inpvec) {
      setOut(inpvec) ;
   } 

   //--------------------------------------------------
   // get net output

   void BpLayer::getNetOutput(ValVector & outvec) {
      getOut(outvec) ;
   }

   //--------------------------------------------------
   // propFore

   void BpLayer::propFore(const ValVector & inpvec, const Val & temp) {
      return setNetInput(inpvec) ;
   } ;

   void BpLayer::propFore(const Val & temp) {
      switch(ftype) {
	 case FT_linear : return propForeLinear(temp) ;
	 case FT_sigmoid : return propForeSigmoid(temp) ;
	 case FT_bigmoid : return propForeBigmoid(temp) ;
	 default : throw 0 ;
      }
   } ;

   void BpLayer::propForeLinear(const Val & temp) {
      collectInput(True) ;
      activateByLinear(temp) ;
   } ;

   void BpLayer::propForeSigmoid(const Val & temp) {
      collectInput(True) ;
      activateBySigmoid(temp) ;
   } ;

   void BpLayer::propForeBigmoid(const Val & temp) {
      collectInput(True) ;
      activateByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // clear input to same value

   void BpLayer::clearInput(const Val & value) {
      for(Iterator u = begin() ; u < end() ; u++) (*u).inp = value ; 
   } 

   //--------------------------------------------------
   // set input vector by threshold

   void BpLayer::clearInputByThreshold() {
      for(Iterator u = begin() ; u < end() ; u++) {
	 (*u).inp = (*u).thr ;
      }
   } 

   //--------------------------------------------------
   // collect input

   void BpLayer::collectInput(const Bool clearp) {
      if(clearp) clearInputByThreshold() ;

      for(LLinkList::Iterator lnk = from.begin() ; lnk < from.end() ; lnk++) 
	 (*lnk)->propFore() ;
   }
      
   //--------------------------------------------------
   // activate

   void BpLayer::activateByLinear(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateByLinear(temp) ;
   }

   void BpLayer::activateBySigmoid(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateBySigmoid(temp) ;
   }

   void BpLayer::activateByBigmoid(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).activateByBigmoid(temp) ;
   }
   
   //----------------------------------------------------------------------
   // Backward processing

   //--------------------------------------------------
   // calc net output error (for output layer)

   void BpLayer::propBack(const ValVector & tchvec, const Val & temp) {
      switch(ftype) {
	 case FT_linear : return propBackLinear(tchvec,temp) ;
	 case FT_sigmoid : return propBackSigmoid(tchvec,temp) ;
	 case FT_bigmoid : return propBackBigmoid(tchvec,temp) ;
	 default : throw 0 ;
      }
   }

   void BpLayer::propBackLinear(const ValVector & tchvec,
				const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaByLinear(temp) ;
   }

   void BpLayer::propBackSigmoid(const ValVector & tchvec,
				const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaBySigmoid(temp) ;
   } ;

   void BpLayer::propBackBigmoid(const ValVector & tchvec,
				 const Val & temp) {
      calcNetError(tchvec) ;
      calcDeltaByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // prop backward (for hidden layer)

   void BpLayer::propBack(const Val & temp) {
      switch(ftype) {
	 case FT_linear : return propBackLinear(temp) ;
	 case FT_sigmoid : return propBackSigmoid(temp) ;
	 case FT_bigmoid : return propBackBigmoid(temp) ;
	 default : throw 0 ;
      }
   }

   void BpLayer::propBackLinear(const Val & temp) {
      collectDelta() ;
      calcDeltaByLinear(temp) ;
   } ;

   void BpLayer::propBackSigmoid(const Val & temp) {
      collectDelta() ;
      calcDeltaBySigmoid(temp) ;
   } ;

   void BpLayer::propBackBigmoid(const Val & temp) {
      collectDelta() ;
      calcDeltaByBigmoid(temp) ;
   } ;

   //--------------------------------------------------
   // set net output error

   void BpLayer::calcNetError(const ValVector & tchvec) {
      Iterator u = begin() ;
      for(ValVector::C_Iterator tch = tchvec.begin() ;
	  tch < tchvec.end() && u < end() ;
	  tch++, u++) {
	 (*u).calcError(*tch) ;
      }
   } 

   //--------------------------------------------------
   // clear Error
   void BpLayer::clearError() {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).err = 0.0 ;
   } 

   //--------------------------------------------------
   // sum of error
   Val BpLayer::sumError() const {
      Val sqerr = 0.0 ;
      for(C_Iterator u = begin() ; u < end() ; u++) {
	 sqerr += (*u).sqerr() ;
      }
      return sqerr ;
   } 

   //--------------------------------------------------
   // collect delta-in

   void BpLayer::collectDelta() {
      clearError() ;
      for(LLinkList::Iterator lnk = to.begin() ; lnk < to.end() ; lnk++)
	 (*lnk)->propBack() ;
   }
      
   //--------------------------------------------------
   // calc delta

   void BpLayer::calcDeltaByLinear(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaByLinear(temp) ;
   } 

   void BpLayer::calcDeltaBySigmoid(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaBySigmoid(temp) ;
   } 
   void BpLayer::calcDeltaByBigmoid(const Val & temp) {
      for(Iterator u = begin() ; u < end() ; u++) 
	 (*u).calcDeltaByBigmoid(temp) ;
   } 

   //----------------------------------------------------------------------
   // learn

   Val BpLayer::learn(const Val & eps, const Val & alpha) {
      Val dsum = 0.0 ;
      for(Iterator u = begin() ; u < end() ; u++) {
	 Val & dh = (*u).dh ;
	 dh *= alpha ;
	 dh += eps * (*u).dlt ;
	 dsum += sq(dh) ;
	 (*u).thr += dh ;
      }
      return dsum ;
   }

   //----------------------------------------------------------------------
   // adjust output by error value

   void BpLayer::adjustOutputByError(const Val & eps) {
      for(Iterator u = begin() ; u < end() ; u++) {
	 (*u).out += (eps * (*u).err) ;
      }
   } 

   //======================================================================
   // Preparation

   //----------------------------------------------------------------------
   // add from/to link information

   Bool BpLayer::addFromLink(BpLLink * lnk) { 
      if(isNull(from.findPtr(lnk))) {
	 from.push(lnk) ; 
	 return True ;
      } else {
	 return False ;
      }
   } ;      

   Bool BpLayer::addToLink(BpLLink * lnk) {
      if(isNull(to.findPtr(lnk))) {
	 to.push(lnk) ; 
	 return True ;
      } else {
	 return False ;
      } 
   } ;

   //----------------------------------------------------------------------
   // randomize threshold

   void BpLayer::randomizeThreshold(const Val & minval, const Val & maxval) {
      randomizeByTime() ;
      for(Iterator u = begin() ; u < end() ; u++) {
	 (*u).thr = fltRand(minval,maxval) ;
      }
   }

   //----------------------------------------------------------------------
   // access to each vector

#define SETVEC_BODY(valvec,val) {\
      ValVector::C_Iterator val = valvec.begin() ;\
      for(Iterator u = begin() ;\
	  u < end() && val < valvec.end() ; u++, val++)\
	 (*u).val = *val ;\
}
#define GETVEC_BODY(valvec,val) {\
      ValVector::Iterator val = valvec.begin() ;\
      for(C_Iterator u = begin() ;\
	  u < end() && val < valvec.end() ; u++, val++)\
	 *val = (*u).val ;\
}
	
   
   void BpLayer::setInp(const ValVector & inpvec) { SETVEC_BODY(inpvec,inp) }
   void BpLayer::getInp(ValVector & inpvec) const { GETVEC_BODY(inpvec,inp) }

   void BpLayer::setOut(const ValVector & outvec) { SETVEC_BODY(outvec,out) }
   void BpLayer::getOut(ValVector & outvec) const { GETVEC_BODY(outvec,out) }

   void BpLayer::setThr(const ValVector & thrvec) { SETVEC_BODY(thrvec,thr) }
   void BpLayer::getThr(ValVector & thrvec) const { GETVEC_BODY(thrvec,thr) }

   void BpLayer::setErr(const ValVector & errvec) { SETVEC_BODY(errvec,err) }
   void BpLayer::getErr(ValVector & errvec) const { GETVEC_BODY(errvec,err) }

   void BpLayer::setDlt(const ValVector & dltvec) { SETVEC_BODY(dltvec,dlt) }
   void BpLayer::getDlt(ValVector & dltvec) const { GETVEC_BODY(dltvec,dlt) }

   void BpLayer::setDh(const ValVector & dhvec) { SETVEC_BODY(dhvec,dh) }
   void BpLayer::getDh(ValVector & dhvec) const { GETVEC_BODY(dhvec,dh) }


   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void BpLayer::save(ostream & ostr,
		      const Bool & newlinep,
		      const SubString & indent) const {
      SubString indent2 = indent + "  " ; // for each param
      //SubString indent3 = indent2 + "  " ; // for each inner strForm

      StrForm::writeBeginTerm(ostr,sformTermHead) ;

      StrForm::writeAssocRole(ostr,sformNameLbl) ; 
      StrForm::writeString(ostr,name) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,sformSizeLbl) ; 
      ostr << n << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,sformFuncTypeLbl) ; 
      ostr << ftypeName(ftype) << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,sformLrnRateLbl) ; 
      ostr << NNN_VAL_FORMAT << lrate << ITK_SST_TERM_SEP ;

      ostr << endl << indent2 ;
      StrForm::writeAssocRole(ostr,sformUnitLbl) ; {
	 SubString indent4 = indent2 + "        " ;
	 StrForm::writeBeginTerm(ostr) ;
	 StrForm::writeBeginList(ostr) ;
	 for(Index i = 0 ; i < n ; i++) {
	    if(i > 0) { StrForm::writeArgSep(ostr) ; } 
	    if(i > 0 && i%4 == 0) ostr << endl << indent4 ; // indent
	    unit(i).save(ostr) ;
	 }
	 StrForm::writeEndTerm(ostr) ;
      }
      //ostr << endl << indent ;
      StrForm::writeEndTerm(ostr) ;
      if(newlinep) ostr << endl ;
   } ;
   
   //----------------------------------------------------------------------
   // load

   void BpLayer::scan(StrForm & sform) {
      if(!sform.isTerm(sformTermHead)) {
	 ITK_ERR("Illegal BpLayer StrForm:" << sform) ;
	 throw 0 ;
      } 

      name = sform.assoc(sformNameLbl)->cvtSString() ;

      ftype = ftypeByName(sform.assoc(sformFuncTypeLbl)->cvtSString()) ;

      resize(sform.assoc(sformSizeLbl)->cvtUInt()) ;

      sform.assocCvtFlt(sformLrnRateLbl,lrate) ;

      StrForm * unitlist = sform.assoc(sformUnitLbl) ;
      if(!isNull(unitlist)) {
	 for(Index i = 0 ; i < n ; i++) {
	    unit(i).scan(*(unitlist->nthListArg(i))) ;
	 } 
      }

      initInfo = ITK_NULLPTR ;
      sform.assocCvtStrFormPtr(sformInitLbl,initInfo) ;
      if(!isNull(initInfo)) {
	 if(initInfo->isTerm("rand")) {
	    randomizeThreshold(initInfo->nthListArg(0)->cvtFlt(),
			       initInfo->nthListArg(1)->cvtFlt()) ;
	 } else {
	    ITK_ERR("illegal init value specifier" << *initInfo) ;
	    throw 0 ;
	 }
	 initInfo = ITK_NULLPTR ;
      }

   } ;



} ;
