// -*- Mode:C++ -*-
//Header:
//File: StrForm.h
//Author: NODA, Itsuki
//Date: 1999/12/23
//

//ModifyHistory:
// 1999/12/23: Start to create this file
//EndModifyHistory:

//StrForm Syntax:
//	<Form> ::= <Atom> | <Term>
//	<Atom> ::= <Number_Symbol_Atom> | <String>
//	<Number_Symbol_Atom> ::= that does not include separators
//	<Symbol> ::= `"' (any char string except `"'(escape = `/')) `"'
//	<Term> ::= `#' <Head> `[' <AssocArgs> [ `|' <ListArgs> ] `]'

#ifndef _Itk_StrForm_h_
#define _Itk_StrForm_h_
////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include "itk/itk.h"


namespace Itk {

//======================================================================
// special charactors

#define ITK_SST_NONATOMCHAR	"=#[](){},|\" \n\t"

#define ITK_SST_TERM_PREFIX	'#'
#define ITK_SST_TERM_BEGINS	"[({"
#define ITK_SST_TERM_ENDS	"])}"
#define ITK_SST_TERM_BEGIN	'['
#define ITK_SST_TERM_END	']'
#define ITK_SST_TERM_SEP	','
#define ITK_SST_TERM_LISTARGSEP	'|'
#define ITK_SST_TERM_ASSOCOP	'='

#define ITK_SST_STRING_BEGIN	'"'
#define ITK_SST_STRING_END	'"'
#define ITK_SST_STRING_ESCAPE	'\\'

#define ITK_SST_COMMENT_BEGIN	'%'
#define ITK_SST_COMMENT_END	'\n'

#define ITK_SST_VAR_HEAD	"$"

#define ITK_SST_INITBUFSIZE	1024

//======================================================================
// class StrForm

   class StrForm : public WithDescriber {
	 //--------------------------------------------------
	 // RoleValue pair
      public:
	 class AssocArg : public WithDescriber {
	       //----------------------------------------
	       // member
	    public:
	       SubString role ;
	       StrForm * value ;
	       AssocArg * next ;

	       //----------------------------------------
	       // constructor
	    public:
	       AssocArg() : role(), value(ITK_NULLPTR), next(ITK_NULLPTR) {};

	       //----------------------------------------
	       // descructor
	    public:
	       virtual ~AssocArg() { destroy() ; } ;
	       void destroy() {
		  if(!isNull(value)) delete value ;
		  if(!isNull(next)) next->destroy() ;
	       } ;

	       //----------------------------------------
	       // length
	    public:
	       UInt length() const {
		  if(isNull(this)) return 0 ;
		  else             return 1 + next->length() ; 
	       } ;
	       
	       //----------------------------------------
	       // nth
	    public:
	       StrForm * nth(UInt n) {
		  AssocArg * arg = nthPos(n) ;
		  if(isNull(arg)) return ITK_NULLPTR ;
		  else return arg->value ;
	       } ;
	       AssocArg * nthPos(UInt n) {
		  if(isNull(this)) return ITK_NULLPTR ;
		  else if(n == 0)  return this ;
		  else             return next->nthPos(n-1) ;
	       } ;

	       //----------------------------------------
	       // assoc
	    public:
	       StrForm * assoc(const SubString& r) {
		  AssocArg * arg = assocPos(r) ;
		  if(isNull(arg)) return ITK_NULLPTR ;
		  else return arg->value ;
	       } ;
	    public:
	       AssocArg * assocPos(const SubString& r) {
		  if(isNull(this)) return ITK_NULLPTR ;
		  else if(role==r) return this ;
		  else             return next->assocPos(r) ;
	       } ;

	       //----------------------------------------
	       // describe
	    public:
	       virtual void describe(ostream & ostr,
				     const Bool detailp = True) const {
		  ostr << "#AssocArg[" << role << ":" ;
		  if(isNull(value)) ostr << "(null)" ;
		  else              ostr << *value ;
		  ostr << "]" ;
		  if(detailp) {
		     if(!isNull(next)) {
			ostr << ", " ;
			next->describe(ostr,detailp) ;
		     }
		     ostr << endl ;
		  } ;
	       } ;
	 } ;

	 //--------------------------------------------------
	 // Form Type
      public:
	 enum Type { Type_Atom, Type_Term } ;
	 
	 //--------------------------------------------------
	 // member
      public:
	 SubString str ;
	 Type type ;
	 SubString head ;
	 AssocArg * assocArg ;
	 StrForm * listArg ;
	 StrForm * next ;	/* used in listArg */

	 //--------------------------------------------------
	 // constructor
      public:
	 StrForm() : 
	    str(), head(), 
	    assocArg(ITK_NULLPTR), listArg(ITK_NULLPTR),
	    next(ITK_NULLPTR) {
	 } ;

	 //--------------------------------------------------
	 // destructor
      public:
	 virtual ~StrForm() { destroy() ; } ;
	 
	 void destroy(Bool toplevelp = False) {
	    if(!isNull(assocArg)) delete assocArg ;
	    if(!isNull(listArg)) listArg->destroy() ;
	    if(!isNull(next)) next->destroy() ;
	    if(toplevelp) {
	       delete str ;
	       delete this ;
	    }
	 } ;

	 //--------------------------------------------------
	 // prescan utils
      public:
	 static void chkEndError(istream & istr) ;
	 static UInt skipSpace(istream & istr) ;
	 static UInt skipComment(istream & istr) ;
	 static UInt skipSpaceComment(istream & istr) ;
		  
	 //--------------------------------------------------
	 // prescan 
      public:
	 static Buffer prescan(istream & istr) ;
	 static Bool prescanTop(Buffer & buf, istream & istr) ;
	 static Bool prescanAtom(Buffer & buf, istream & istr) ;
	 static Bool prescanString(Buffer & buf, istream & istr) ;
	 static Bool prescanTerm(Buffer & buf, istream & istr) ;
	 static Bool prescanTermAssocArg(Buffer & buf, istream & istr) ;
	 static Bool prescanTermAssocArgSub(Buffer & buf, istream & istr) ;
	 static Bool prescanTermListArg(Buffer & buf, istream & istr) ;
	 static Bool prescanTermListArgSub(Buffer & buf, istream & istr) ;

	 //--------------------------------------------------
	 // scan utils
      public:
	 static void chkEndError(Buffer & buf) ;
	 static UInt skipSpace(Buffer & buf) ;
	 static UInt skipComment(Buffer & buf) ;
	 static UInt skipSpaceComment(Buffer & buf) ;
		  
	 //--------------------------------------------------
	 // scan 
      public:
	 static StrForm * scan(istream & istr) {
	    Buffer buf = prescan(istr) ;
	    return scan(buf) ;
	 } ;

	 static StrForm * scan(Buffer & buf) ;

	 static StrForm * scanTop(Buffer & buf) ;
	 static StrForm * scanAtom(Buffer & buf) ;
	 static StrForm * scanString(Buffer & buf) ;
	 static StrForm * scanTerm(Buffer & buf) ;
	 static AssocArg * scanTermAssocArg(Buffer & buf) ;
	 static AssocArg * scanTermAssocArgSub(Buffer & buf) ;
	 static StrForm * scanTermListArg(Buffer & buf) ;
	 static StrForm * scanTermListArgSub(Buffer & buf) ;

	 //--------------------------------------------------
	 // type check
      public:
	 Bool isAtom() const { return !isNull(this) && type == Type_Atom ; } ;
	 Bool isTerm() const { return !isNull(this) && type == Type_Term ; } ;
	 Bool isString() const { 
	    return isAtom() && str.first() == ITK_SST_STRING_BEGIN ;
	 } ;
	 Bool isPureAtom() const { 
	    return isAtom() && str.first() != ITK_SST_STRING_BEGIN ;
	 } ;
	 Bool hasAssocArg() const { return isTerm() && !isNull(assocArg) ; } ;
	 Bool hasListArg() const { return isTerm() && !isNull(listArg) ; } ;
	 Bool isAnonymousTerm() const {
	    return isTerm() && head.nullp() ; 
	 } ;
	 Bool isTerm(const SubString & hd) const {
	    return isTerm() && head == hd ;
	 } ;

	 //--------------------------------------------------
	 // convert
      public:
	 Int cvtInt() const {
	    if(!isAtom()) { 
	       ITK_ERR("Illegal Conversion to Int:" << *this) ;
	       throw 0 ;
	    }
	    return strtol(str,ITK_NULLPTR,0) ;
	 } ;

      public:
	 UInt cvtUInt() const {
	    if(!isAtom()) {
	       ITK_ERR("Illegal Conversion to UInt:" << *this) ;
	       throw 0 ;
	    }
	    return strtoul(str,ITK_NULLPTR,0) ;
	 } ;

      public:
	 Flt cvtFlt() const {
	    if(!isAtom()) {
	       ITK_ERR("Illegal Conversion to Flt:" << *this) ;
	       throw 0 ;
	    }
	    return strtod(str,ITK_NULLPTR) ;
	 } ;

      public:
	 SubString cvtSString(Bool dupp = True) const {
	    if(!isAtom()) {
	       ITK_ERR("Illegal Conversion to Flt:" << *this) ;
	       throw 0 ;
	    }
	    if(isString()) {
	       if(dupp) return head.dup() ;
	       else return head ;
	    } else {
	       if(dupp) return str.dup() ;
	       else return str ;
	    }
	 } ;
	 
	 //--------------------------------------------------
	 // length of argument
      public:
	 UInt lengthAssocArg() const { return assocArg->length() ; } ;
	 UInt lengthListArg() const {
	    UInt l = 0 ;
	    for(StrForm * arg = listArg ; !isNull(arg) ; arg = arg->next ) 
	       l++ ;
	    return l ;
	 } ;

	 //--------------------------------------------------
	 // argument opeartion
      public:
	 StrForm * assoc(const SubString & role) const {
	    return assocArg->assoc(role) ;
	 } ;
	 AssocArg * assocPos(const SubString & role) const {
	    return assocArg->assocPos(role) ;
	 } ;

	 StrForm * nthAssocArg(UInt n) const {
	    return assocArg->nth(n) ;
	 } ;
	 AssocArg * assocPos(UInt n) const {
	    return assocArg->nthPos(n) ;
	 } ;

	 StrForm * nthListArg(UInt n) const {
	    UInt l = 0 ;
	    for(StrForm * arg = listArg ; !isNull(arg) ; arg = arg->next) {
	       if(l == n) return arg ;
	       l++ ;
	    } ;
	    return ITK_NULLPTR ;
	 } ;
	    
	 //--------------------------------------------------
	 // assoc & convert if exist
      public:
	 Bool assocCvtInt(const SubString & role, Int & value) const {
	    const StrForm * s = assoc(role) ;
	    if(isNull(s)) return False ;
	    value = s->cvtInt() ;
	    return True ;
	 } ;

      public:
	 Bool assocCvtUInt(const SubString & role, UInt & value) const {
	    const StrForm * s = assoc(role) ;
	    if(isNull(s)) return False ;
	    value = s->cvtUInt() ;
	    return True ;
	 } ;

      public:
	 Bool assocCvtFlt(const SubString & role, Flt & value) const {
	    const StrForm * s = assoc(role) ;
	    if(isNull(s)) return False ;
	    value = s->cvtFlt() ;
	    return True ;
	 } ;

      public:
	 Bool assocCvtSString(const SubString & role, SString & value,
			      Bool dupp = True) const {
	    const StrForm * s = assoc(role) ;
	    if(isNull(s)) return False ;
	    value = s->cvtSString(dupp) ;
	    return True ;
	 } ;

      public:
	 Bool assocCvtStrFormPtr(const SubString & role, 
				 StrForm *& value) const {
	    const StrForm * s = assoc(role) ;
	    if(isNull(s)) return False ;
	    value = s->dup() ;
	    return True ;
	 } ;

	 //--------------------------------------------------
	 // assoc as var table (experimental)
      public:
	 StrForm * getVarVal(const SubString & var) {
	    if(isNull(this)) {
	       ITK_ERR("StrForm::getVarVal(): var table is NULL"
		       << " :var=" << var) ;
	       throw 0 ;
	    }
	    
	    StrForm * val = assoc(var) ;
	    if(isNull(val)) {
	       ITK_WRN(ITK_HERE << "can not find var '" << var 
		       << "' in " << *this) ;
	       return ITK_NULLPTR ;
	    } else {
	       return val ;
	    }
	 } ;

	 StrForm * getVarVal(const StrForm & var) {
	    return getVarVal(var.cvtSString(False)) ;
	 } ;
	 
	 StrForm * getVarVal(const StrForm * var) {
	    return getVarVal(*var) ;
	 } ;


	 //--------------------------------------------------
	 // duplicate 
      public:
	 StrForm * dup() const {
	    if(isNull(this)) return ITK_NULLPTR ;

	    Buffer b ; b.set(str.dup()) ;
	    return scan(b) ;
	 } ;
	 
	 //--------------------------------------------------
	 // output utility
      public:
	 static void writeBeginTerm(ostream& ostr) {
	    ostr << ITK_SST_TERM_PREFIX << ITK_SST_TERM_BEGIN ;
	 } ;
	 static void writeBeginTerm(ostream& ostr, const SubString & head) {
	    ostr << ITK_SST_TERM_PREFIX << head << ITK_SST_TERM_BEGIN ;
	 } ;
	 static void writeEndTerm(ostream& ostr) {
	    ostr << ITK_SST_TERM_END ;
	 } ;
	 static void writeAssocRole(ostream& ostr, const SubString & role) {
	    ostr << role << ITK_SST_TERM_ASSOCOP ;
	 } ;
	 static void writeArgSep(ostream& ostr) {
	    ostr << ITK_SST_TERM_SEP ;
	 } ;
	 static void writeBeginList(ostream& ostr) {
	    ostr << ITK_SST_TERM_LISTARGSEP ;
	 } ;
	 static void writeString(ostream& ostr, const SubString & str) {
	    ostr << ITK_SST_STRING_BEGIN << str << ITK_SST_STRING_END ;
	 } ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const ;
   } ;
}

////////////////////////////////////////////////////////////////////////
#endif
