// -*- Mode:C++ -*-
//Header:
//File: VBpLLink.cc
//Author: NODA, Itsuki
//Date: 2000/01/13
//

//ModifyHistory:
// 2000/01/13: Start to create this file
//EndModifyHistory:

#include "VBpLLink.h"

namespace Nnn {

   //======================================================================
   // Execition

   //----------------------------------------------------------------------
   // forward propagation

   //--------------------------------------------------
   // propagate forward

   void VBpLLink::propFore() {
      for(Index i = 0 ; i < n ; i++) {
	 Val & dst = to->inp(i) ;
	 for(Index j = 0 ; j < m ; j++) {
	    dst += from->out(j) * w(i,j) ;
	 }
      }
   }

   //--------------------------------------------------
   // propagate backward

   void VBpLLink::propBack() {
      for(Index j = 0 ; j < m ; j++) {
	 Val & dst = from->err(j) ;
	 for(Index i = 0 ; i < n ; i++) {
	    dst += to->dlt(i) * w(i,j) ;
	 }
	 dst *= from->erate ;
      }
   }

   //--------------------------------------------------
   // learn

   Val VBpLLink::learn(const Val & eps, const Val & alpha) {
      if(lrate == 0.0) return 0.0 ;	// for speed-up
      
      Val dsum = 0.0 ;

      Val e = eps * lrate ;

      for(Index i = 0 ; i < n ; i++) {
	 for(Index j = 0 ; j < m ; j++) {
	    Val & dwv = dw(i,j) ;
	    Val & wv = w(i,j) ;
	    dwv *= alpha ;
	    dwv += e * to->dlt(i) * from->out(j) ;
	    dwv -= decay * wv ;
	    wv += dwv ;
	    dsum += sq(dwv) ;
	    // for debug to avoid overflow
	    //if(Abs(dwv) > TooBigChange) NNN_XPT(0,"too big dw",this) ;
	 }
      }

     if(norm > 0.0)
	normalizeWeight(norm) ;

      return dsum ;
   } ;

   //======================================================================
   // copy

   //----------------------------------------------------------------------
   // duplicate

   VBpLLink & VBpLLink::duplicate(const Bool sharedp = False,
				  const VBpLayerTable * lytab) const {
      VBpLLink * lnk = new VBpLLink() ;
      copyTo(*lnk,sharedp,True,lytab) ;
      return *lnk ;
   } ;

   //----------------------------------------------------------------------
   // copyFrom


   VBpLLink & VBpLLink::copyFrom(const VBpLLink & src, 
				 const Bool sharedp,
				 const Bool allparamp, 
				 const VBpLayerTable * lytab) {
      if(allparamp) {
	 fromname.copy(src.fromname,True) ;
	 toname.copy(src.toname,True) ;
	 resize(src.n, src.m, False) ;

	 copyMiscParamFrom(src) ;

	 if(!isNull(lytab)) {
	    fixFromTo(*lytab) ;
	    registerFromTo() ;
	 }
      }

      copyWeightFrom(src,sharedp) ;
      
      return *this ;
   } ;

   void VBpLLink::copyMiscParamFrom(const VBpLLink & src) {
      lrate = src.lrate ;
      norm = src.norm ;
      decay = src.decay ;
      initInfo = src.initInfo->dup() ;
      wEquivInfo = src.wEquivInfo->dup() ;
      wEquivLink = ITK_NULLPTR ; 		/* ??? */
   } ;

   void VBpLLink::copyWeightFrom(const VBpLLink & src, 
				 const Bool sharedp) {
      if(sharedp) {
	 _w = src._w ;
      } else {
	 _w->copyFrom(*src._w) ;
      } 
      _dw->copyFrom(*src._dw) ;
   } ;
   

   //======================================================================
   // Management

   //----------------------------------------------------------------------
   // fixFromTo

   Bool VBpLLink::fixFromTo(const VBpLayerTable & layerTable) {
      from = layerTable.find(fromname) ;
      to = layerTable.find(toname) ;
      if(!isNull(from) && !isNull(to)) {
	 resize(to->n,from->n) ;
	 return True ;
      } else {
	 return False ;
      } 
   } 

   //----------------------------------------------------------------------
   // registerFromTo (register link information into From/To layer)

   void VBpLLink::registerFromTo() {
      if(!isNull(from) && !isNull(to)) {
	 from->addToLink(this) ;
	 to->addFromLink(this) ;
      } else {
	 ITK_ERR("can do 'VBpLLink::registerFromTo(): from=" << from 
		 << ", to=" << to) ;
	 throw 0 ;
      } 
      return ;
   } 

   //----------------------------------------------------------------------
   // randomize weight
   
   void VBpLLink::randomizeWeight(const Val & minval, const Val & maxval) {
      randomizeByTime() ;
      for(Index i = 0 ; i < n ; i++) {
	 for(Index j = 0 ; j < m ; j++) {
	    w(i,j) = fltRand(minval,maxval) ;
	 }
      }
   } 

   //----------------------------------------------------------------------
   // normalize weight
   
   void VBpLLink::normalizeWeight(const Val & nm) {
      for(Index i = 0 ; i < n ; i++) {
	 Val r = 0.0 ;
	 for(Index j = 0 ; j < m ; j++) {
	    r += sq(w(i,j)) ;
	 }
	 if(r > nm) {
	    Val a = nm/r ;
	    for(Index j = 0 ; j < m ; j++) {
	       w(i,j) *= a ;
	    }
	 }
      }
   }


   //----------------------------------------------------------------------
   // fixEquiv

   Bool VBpLLink::fixEquiv(const VBpLLinkTable & linkTable) {
      if(isNull(wEquivInfo)) return False ;

      StrForm * objinfo = wEquivInfo->assoc(LblObj) ;
      if(isNull(objinfo)) {
	 ITK_ERR("Illegal StrForm for equiv info:" << *wEquivInfo) ;
	 throw 0 ;
      }

      // Here, objinfo shouldbe "#link[...]"

      SString f, t ;
      wEquivInfo->assocCvtSString(LblFrom,f) ;
      wEquivInfo->assocCvtSString(LblTo,t) ;

      VBpLLink * lnk = linkTable.find(f,t) ;
      if(isNull(lnk)) {
	 ITK_ERR("Unknown equiv link:" << f << "->" << t) ;
	 throw 0 ;
      } ;
      wEquivLink = lnk ;

      SString vecname ;
      wEquivInfo->assocCvtSString(LblVec,vecname,False) ;
      _w = &(wEquivLink->vec(vecname)) ;

      return True ;
   } 

   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void VBpLLink::save(ostream & ostr,
		       SaveLevel slevel,
		       const Bool & newlinep,
		       const SubString & indent) const {
      SubString indent2 = indent + "  " ;

      StrForm::writeBeginTerm(ostr,LblTermHead) ;

      StrForm::writeAssocRole(ostr,LblFrom) ; 
      StrForm::writeString(ostr,fromname) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,LblTo) ; 
      StrForm::writeString(ostr,toname) ;
      StrForm::writeArgSep(ostr) ;

      StrForm::writeAssocRole(ostr,LblToSize) ; 
      ostr << n << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblFromSize) ; 
      ostr << m << ITK_SST_TERM_SEP ;

      StrForm::writeAssocRole(ostr,LblLrnRate) ; 
      ostr << NNN_VAL_FORMAT << lrate << ITK_SST_TERM_SEP ;

      if(norm > 0) {
	 StrForm::writeAssocRole(ostr,LblNorm) ; 
	 ostr << NNN_VAL_FORMAT << norm << ITK_SST_TERM_SEP ;
      }

      if(decay > 0) {
	 ostr << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblDecay) ; 
	 ostr << NNN_VAL_FORM(20,16) << decay << ITK_SST_TERM_SEP ;
      }

      // save weight matrix
      ostr << endl << indent2 ;
      StrForm::writeAssocRole(ostr,LblW) ; {
	 if(!isNull(wEquivInfo)) { /* equiv case */
	    ostr << wEquivInfo->str ;
	 } else {                  /* normal case */
	    _w->saveAsMatrix(ostr,m,indent2 + "  ") ;
	 }
      } ;
      
      // save dw
      if(slevel == SL_RunState || slevel == SL_AllState) {
	 ostr << ITK_SST_TERM_SEP << endl << indent2 ;
	 StrForm::writeAssocRole(ostr,LblDw) ; {
	    if(!isNull(wEquivInfo)) { /* equiv case */
	       ostr << wEquivInfo->str ;
	    } else {                  /* normal case */
	       _dw->saveAsMatrix(ostr,m,indent2 + "  ") ;
	    }
	 } ;
      }

      StrForm::writeEndTerm(ostr) ;
      if(newlinep) ostr << endl ;
   }
   
   //----------------------------------------------------------------------
   // scan

   void VBpLLink::scan(StrForm & sform, 
		       VBpLayerTable * layerTable,
		       VBpLLinkTable * linkTable) {
      if(!sform.isTerm(LblTermHead)) {
	 ITK_ERR("Illegal VBpLLink StrForm:" << sform) ;
	 throw 0 ;
      } 

      sform.assocCvtSString(LblFrom,fromname) ;
      sform.assocCvtSString(LblTo,toname) ;
      if(!isNull(layerTable)) {
	 fixFromTo(*layerTable) ;
      } ;

      UInt sizeN, sizeM ;
      if(sform.assocCvtUInt(LblToSize,sizeN) &&
	 sform.assocCvtUInt(LblFromSize,sizeM)) {
	 resize(sizeN,sizeM) ;
      } else if (!isNull(from) && !isNull(to)) {
	 resize(to->n, from->n) ;
      } 

      sform.assocCvtFlt(LblLrnRate,lrate) ;

      if(!sform.assocCvtFlt(LblNorm,norm)) { norm = 0.0 ; } ;
      if(!sform.assocCvtFlt(LblDecay,decay)) { decay = 0.0 ; } ;

      StrForm * winfo = sform.assoc(LblW) ;
      
      if(isNull(winfo)) {
	 /* ??? */
      } else if(winfo->isTerm(ValVector::LblTermHead)) { /* ValVector */
	 _w->scan(*winfo) ;
      } else if(winfo->isTerm(LblEquiv)) { /* equiv */
	 wEquivInfo = winfo->dup() ;
	 if(!isNull(linkTable)) {
	    fixEquiv(*linkTable) ;
	 }
      } else if(winfo->isTerm(LblRand)) { /* rand */
	 initInfo = winfo->dup() ;
	 randomizeWeight(initInfo->nthListArg(0)->cvtFlt(),
			 initInfo->nthListArg(1)->cvtFlt()) ;
	 initInfo = ITK_NULLPTR ;
      } else {
	    ITK_ERR("illegal init value specifier:" << *initInfo) ;
	    throw 0 ;
      } 
   }
}









