// -*- Mode:C++ -*-
//Header:
//File: VBpLayer.h
//Author: NODA, Itsuki
//Date: 2000/01/12
//

//ModifyHistory:
// 2000/01/12: Start to create this file
//EndModifyHistory:

#ifndef _Nnn_VBpLayer_h_
#define _Nnn_VBpLayer_h_
////////////////////////////////////////////////////////////////////////

#include "n3btype.h"
#include "n3utility.h"
#include "StrForm.h"
#include "ValVector.h"
#include "VBpUnitWrap.h"
#include "VBpLinkWrap.h"

namespace Nnn {

   using namespace Itk ;

   class VBpLLink ;
   class VBpLayerTable ;

   //======================================================================
   // BpLayer

   class VBpLayer : public WithDescriber {

	 //--------------------------------------------------
	 // labels for save/load/scan
      public:
	 static char * const LblTermHead = "VBpLayer" ; 
	 static char * const LblName = "name" ; 
	 static char * const LblSize = "n" ; 
	 //static char * const LblUnit = "unit" ; 
	 static char * const LblFuncType = "ftype" ; 
	 static char * const LblLrnRate = "lrate" ; 
	 static char * const LblErrRate = "erate" ; 
	 static char * const LblSatVal = "satval" ;
	 static char * const LblSatRate = "satrate" ;
	 static char * const LblInpNorm = "inpnorm" ;  
	 static char * const LblEquiv = "equiv" ;
	 static char * const LblInit = "init" ; 
	 static char * const LblRand = "rand" ; 
	 static char * const LblInp = "inp" ;
	 static char * const LblOut = "out" ;
	 static char * const LblThr = "thr" ;
	 static char * const LblErr = "err" ;
	 static char * const LblDlt = "dlt" ;
	 static char * const LblDh = "dh" ;
	 static char * const LblObj = "obj" ;
	 static char * const LblVec = "vec" ;

	 //--------------------------------------------------
	 // typedef
      public:
	 typedef StackT<VBpLLink*> LLinkList ;

	 //--------------------------------------------------
	 // Function Type
      public:
	 enum FuncType { FT_linear, FT_sigmoid, FT_bigmoid } ;
	 static const char * ftypeName(FuncType type) {
	    switch(type) {
	       case FT_linear : return "linear" ;
	       case FT_sigmoid : return "sigmoid" ;
	       case FT_bigmoid : return "bigmoid" ;
	    }
	    return ITK_NULLPTR ;
	 } ;
	 static FuncType ftypeByName(const SString & name) {
	    if(name == "linear") return FT_linear ;
	    if(name == "sigmoid") return FT_sigmoid ;
	    if(name == "bigmoid") return FT_bigmoid ;
	    throw 0 ;
	 } ;

	 //--------------------------------------------------
	 // member
      public:
	 SubString name ;
	 Size n ;
	 FuncType ftype ;
	 Val lrate ;
	 Val erate ;
	 Val satrate ; // rate of saturate stress of net-in
	 Val satval ;  // cutoff threshold value of saturated net-in

	 Val inpnorm ; // threshold to normalize input weight by net-in
	 	       // if inpnorm =< 0.0, do not normailze
	 	       // [not used now]
	 
	 // Val temp ;	// means to use network's temp
	 // Val eps ;	// learning step
	 // Val alpha ;	// ineatia
	 
	 StrForm * initInfo ;

	 LLinkList from ;
	 LLinkList to ;

	 //BpUnitBase * _unit ;

	 ValVector * _inp ;
	 ValVector * _out ;
	 ValVector * _thr ;
	 ValVector * _err ;
	 ValVector * _dlt ;
	 ValVector * _dh ;
	 ValVector * _rinp ;	/* required input */

	 StrForm * thrEquivInfo ;
	 VBpLayer * thrEquivLayer ;

	 //--------------------------------------------------
	 // constructor
      public:
	 VBpLayer() { init(SString()) ; } ;
	 VBpLayer(const SubString & nm, 
		  Size size = 0, 
		  FuncType ft = FT_sigmoid,
		  Bool allocp = True) { 
	    init(nm,size,ft,allocp) ;
	 } ;

	 //--------------------------------------------------
	 // initialize
      public:
	 void init(const SubString & nm, 
		   Size size = 0, 
		   FuncType ft = FT_sigmoid,
		   Bool allocp = True) {
	    name.copy(nm,True) ;
	    ftype = ft ;
	    n = 0 ;
	    if(allocp) {
	       if(size > 0) {
		  resize(size,False) ;
	       } else {
		  assignVector(ITK_NULLPTR, ITK_NULLPTR, ITK_NULLPTR,
			       ITK_NULLPTR, ITK_NULLPTR, ITK_NULLPTR,
			       ITK_NULLPTR) ;
	       }
	    } 
	    lrate = 1.0 ;
	    erate = 1.0 ;
	    satrate = 0.0 ;  // saturate rate
	    satval = 0.0 ;   // saturate cutoff value
	    inpnorm = 0.0 ;  // [not used now]

	    thrEquivInfo = ITK_NULLPTR ;
	    thrEquivLayer = ITK_NULLPTR ;
	 } ;

	 //--------------------------------------------------
	 // assign vectors
      public:
	 void assignVector(ValVector * iv, ValVector * ov, ValVector * hv, 
			   ValVector * ev, ValVector * dv, ValVector * cv,
			   ValVector * riv) {
	    _inp = iv ; _out = ov ; _thr = hv ; 
	    _err = ev ; _dlt = dv ; _dh = cv ;
	    _rinp = riv ;
	 } ;
	 void assignVector(ValVector & iv, ValVector & ov, ValVector & hv, 
			   ValVector & ev, ValVector & dv, ValVector & cv,
			   ValVector & riv) {
	    _inp = &iv ; _out = &ov ; _thr = &hv ; 
	    _err = &ev ; _dlt = &dv ; _dh = &cv ;
	    _rinp = &riv ;
	 } ;
	 
	 //--------------------------------------------------
	 // resize
      public:
	 Bool resize(Size newsize, Bool deletep = True) {
	    Size oldsize = n ;
	    n = newsize ;
	    if(oldsize > newsize) return False ;

	    if(deletep) {
	       delete _inp ; delete _out ; delete _thr ;
	       delete _err ; delete _dlt ; delete _dh ;
	    }
	    assignVector(new ValVector(n), new ValVector(n),
			 new ValVector(n), new ValVector(n),
			 new ValVector(n), new ValVector(n),
			 new ValVector(n)) ;
	    return True ;
	 } ;

	 //--------------------------------------------------
	 // unit(i) : access to each unit
      public:
	 VBpUnitWrap _unit(const Index & i) {
	    return VBpUnitWrap(inp(i),out(i),thr(i),err(i),dlt(i),dh(i)) ;
	 } ;
	 VBpUnitWrap unit(const Index & i ) { return _unit(i) ; } ;
	 const VBpUnitWrap unit(const Index & i) const {
	    return ((VBpLayer*)this)->_unit(i) ;
	 } ;
	       
	 //--------------------------------------------------
	 // access to each value
      public:
	 const Val & inp(const Index & i) const { return inp()[i] ; } ;
	 Val & inp(const Index & i) { return inp()[i] ; } ;

	 const Val & out(const Index & i) const { return out()[i] ; } ;
	 Val & out(const Index & i) { return out()[i] ; } ;

	 const Val & thr(const Index & i) const { return thr()[i] ; } ;
	 Val & thr(const Index & i) { return thr()[i] ; } ;

	 const Val & err(const Index & i) const { return err()[i] ; } ;
	 Val & err(const Index & i) { return err()[i] ; } ;

	 const Val & dlt(const Index & i) const { return dlt()[i] ; } ;
	 Val & dlt(const Index & i) { return dlt()[i] ; } ;

	 const Val & dh(const Index & i) const { return dh()[i] ; } ;
	 Val & dh(const Index & i) { return dh()[i] ; } ;

	 //--------------------------------------------------
	 // access to each vector
      public:
	 const ValVector & inp() const { return *_inp ; } ;
	 ValVector & inp() { return *_inp ; } ;

	 const ValVector & out() const { return *_out ; } ;
	 ValVector & out() { return *_out ; } ;

	 const ValVector & thr() const { return *_thr ; } ;
	 ValVector & thr() { return *_thr ; } ;

	 const ValVector & err() const { return *_err ; } ;
	 ValVector & err() { return *_err ; } ;

	 const ValVector & dlt() const { return *_dlt ; } ;
	 ValVector & dlt() { return *_dlt ; } ;

	 const ValVector & dh() const { return *_dh ; } ;
	 ValVector & dh() { return *_dh ; } ;

	 //--------------------------------------------------
	 // access to special vectors (used in VElman5Lb for adaptation)
      public:
	 const ValVector & rinp() const { return *_rinp ; } ;
	 ValVector & rinp() { return *_rinp ; } ; 
	 
	 const Val & rinp(const Index & i) const { return rinp()[i] ; } ;
	 Val & ainp(const Index & i) { return rinp()[i] ; } ;

	 //--------------------------------------------------
	 // generic access to each vector
      public:
	 const ValVector & vec(const SubString & vecname) const {
	    if(vecname == LblInp) return inp() ;
	    if(vecname == LblOut) return out() ;
	    if(vecname == LblThr) return thr() ;
	    if(vecname == LblErr) return err() ;
	    if(vecname == LblDlt) return dlt() ;
	    if(vecname == LblDh) return dh() ;
	    
	    ITK_ERR("Illeval vector name:" << name) ;
	    throw 0 ;
	 } ;
	 ValVector & vec(const SubString & name) {
	    if(name == LblInp) return inp() ;
	    if(name == LblOut) return out() ;
	    if(name == LblThr) return thr() ;
	    if(name == LblErr) return err() ;
	    if(name == LblDlt) return dlt() ;
	    if(name == LblDh) return dh() ;

	    ITK_ERR("Illeval vector name") ;
	    throw 0 ;
	 } ;
	    
	 //--------------------------------------------------
	 // set/get each vector
      public:
	 void setInp(const ValVector & inpvec) ;
	 void setInp(const ValVector * inpvec) { setInp(*inpvec) ; } ;
	 void getInp(ValVector & inpvec) const ;
	 void getInp(ValVector * inpvec) const { getInp(*inpvec) ; } ;

	 void setOut(const ValVector & outvec) ;
	 void setOut(const ValVector * outvec) { setOut(*outvec) ; } ;
	 void getOut(ValVector & outvec) const ;
	 void getOut(ValVector * outvec) const { getOut(*outvec) ; } ;

	 void setThr(const ValVector & thrvec) ;
	 void setThr(const ValVector * thrvec) { setThr(*thrvec) ; } ;
	 void getThr(ValVector & thrvec) const ;
	 void getThr(ValVector * thrvec) const { getThr(*thrvec) ; } ;

	 void setErr(const ValVector & errvec) ;
	 void setErr(const ValVector * errvec) { setErr(*errvec) ; } ;
	 void getErr(ValVector & errvec) const ;
	 void getErr(ValVector * errvec) const { getErr(*errvec) ; } ;

	 void setDlt(const ValVector & dltvec) ;
	 void setDlt(const ValVector * dltvec) { setDlt(*dltvec) ; } ;
	 void getDlt(ValVector & dltvec) const ;
	 void getDlt(ValVector * dltvec) const { getDlt(*dltvec) ; } ;

	 void setDh(const ValVector & dhvec) ;
	 void setDh(const ValVector * dhvec) { setDh(*dhvec) ; } ;
	 void getDh(ValVector & dhvec) const ;
	 void getDh(ValVector * dhvec) const { getDh(*dhvec) ; } ;

	 //--------------------------------------------------
	 // forward propagation
	 //	propFore(inpvec,temp) : for input layer
	 //	propFore(temp) : for hidden & output layer
      public:
	 void propFore(const ValVector & inpvec, const Val & temp) ;
	 void propFore(const Val & temp) ;
	 void propForeWithMask(const ValVector & inpvec, 
			       const ValVector & mask,
			       const Val & temp) ;

	 void setNetInput(const ValVector & ivec) ;
	 void setNetInputWithMask(const ValVector & ivec,
				  const ValVector & mask) ;
	 void getNetOutput(ValVector & ovec) ;

	 void propForeLinear(const Val & temp) ;
	 void propForeSigmoid(const Val & temp) ;
	 void propForeBigmoid(const Val & temp) ;

	 void clearInput(const Val & value = 0.0) ;
	 void clearInputByThreshold() ;

	 void collectInput(const Bool clearp = True) ;

	 void activateByLinear(const Val & temp) ;
	 //void activateByLinear() { return activateByLinear(temp) ; } ;

	 void activateBySigmoid(const Val & temp) ;
	 //void activateBySigmoid() { return activateBySigmoid(temp) ; } ;

	 void activateByBigmoid(const Val & temp) ;
	 //void activateByBigmoid() { return activateByBigmoid(temp) ; } ;

	 //--------------------------------------------------
	 // backward propagation
	 //	propBack(tchvec,temp): for output layer
	 //	propBack(temp) : for hidden layer
      public:
	 void propBack(const ValVector & tchvec, const Val & temp) ;
	 void propBackWithMask(const ValVector & tchvec, 
			       const ValVector & mask,
			       const Val & temp) ;

	 void propBack(const Val & temp) ;

	 void propBackLinear(const ValVector & tchvec, 
			     const Val & temp) ;
	 void propBackSigmoid(const ValVector & tchvec,
			      const Val & temp) ;
	 void propBackBigmoid(const ValVector & tchvec,
			      const Val & temp) ;

	 void propBackLinear(const Val & temp) ;
	 void propBackSigmoid(const Val & temp) ;
	 void propBackBigmoid(const Val & temp) ;

	 void clearError() ;

	 void calcNetError(const ValVector & tchvec) ;
	 void calcNetErrorWithMask(const ValVector & tchvec,
				   const ValVector & mask) ;
	 void collectDelta() ;
	 void calcDeltaByLinear(const Val & temp) ;
	 void calcDeltaBySigmoid(const Val & temp) ;
	 void calcDeltaByBigmoid(const Val & temp) ;

	 void addSaturateStress() ;

	 void adaptRInpByDelta(const Val & alpha) ;
	 void adaptOutByDelta(const Val & alpha) ;

	 //--------------------------------------------------
	 // normalize weight by netin using inpnorm  [not implemented]
      public:
	 //void normalizeWeightByNetIn() ;

	 //--------------------------------------------------
	 // adjust output by error value
      public:
	 void adjustOutputByError(const Val & eps) ;

	 //--------------------------------------------------
	 // calc sum of error

	 Val sumError() const ;

	 //--------------------------------------------------
	 // learn
      public:
	 Val learn(const Val & eps, const Val & alpha) ;

	 //--------------------------------------------------
	 // add/remove link
      public:
	 Bool addFromLink(VBpLLink * lnk) ;
	 Bool addToLink(VBpLLink * lnk) ;

	 Bool removeFromLink(VBpLLink * lnk) ;
	 Bool removeToLink(VBpLLink * lnk) ;

	 //--------------------------------------------------
	 // randomize threshold
      public:
	 void randomizeThreshold(const Val & minval, const Val & maxval) ;

	 //--------------------------------------------------
	 // fix equiv vectors
      public:
	 Bool fixEquiv(const VBpLayerTable & layerTable) ;

	 //--------------------------------------------------
	 // save 
      public:
	 virtual void save(ostream & ostr,
			   SaveLevel slevel = SL_NetState,
			   const Bool & newlinep = True,
			   const SubString & indent = SubString("")) const ;

	 //--------------------------------------------------
	 // load
      public:
	 virtual void load(istream & istr, 
			   VBpLayerTable * layertable = ITK_NULLPTR) {
	    StrForm * sform = StrForm::scan(istr) ;
	    scan(*sform, layertable) ;
	    delete sform ;
	 } ;

      public:
	 virtual void scan(StrForm & sform,
			   VBpLayerTable * layertable = ITK_NULLPTR) ;

	 //--------------------------------------------------
	 // duplicate, copy
	 // if sharedp == True, then thr is shared
      public:
	 VBpLayer & duplicate(const Bool sharedp = False) const ;

      public:
	 VBpLayer & copyTo(VBpLayer & dest, 
			   const Bool sharedp = False, 
			   const Bool allparamp = False) const {
	    return dest.copyFrom(*this,sharedp,allparamp) ;
	 } ;
	 VBpLayer & copyFrom(const VBpLayer & src, 
			     const Bool sharedp = False,
			     const Bool allparamp = False) ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#VBpLayer[" 
		 << name
		 << "|size=" << n
		 << ", ftype=" << ftypeName(ftype) 
		 << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "  " << "unit= (" ;
	       for(Index i = 0 ; i < n ; i++) {
		  if(i > 0) { ostr << "," << endl << "         " ; } 
		  ostr << unit(i) ;
	       }
	       ostr << ")" << endl ;
	       ostr << "  " << "from= " ; from.describe(ostr) ;
	       ostr << "  " << "to= " ; to.describe(ostr) ;
	    } 
	 } ;

	 //--------------------------------------------------
	 // iterator
      public:
	 /*
	 typedef VBpUnitWrap * Iterator ;
	 typedef const VBpUnitWrap * C_Iterator ;
	 
	 Iterator begin() { return _unit ; } ;
	 C_Iterator begin() const { return _unit ; } ;
	 Iterator end() { return &(_unit[n]) ; } ;
	 C_Iterator end() const { return &(_unit[n]) ; } ;
	 */

	 typedef ValVector::Iterator VIterator ;
	 typedef ValVector::C_Iterator C_VIterator ;
	 typedef ValVector::RevIterator VRevIterator ;
	 typedef ValVector::C_RevIterator C_VRevIterator ;
   } ;

   //======================================================================
   // VBpLayerTable

   class VBpLayerTable : public NameTableT<VBpLayer *> {
      public:
	 Entry * assign(VBpLayer * lyr) {
	    return NameTable::assign(lyr->name, lyr) ;
	 } ;
	 Entry * assign(VBpLayer & lyr) {  return assign(&lyr) ; } ;
   } ;

} ;

#include "VBpLLink.h"

////////////////////////////////////////////////////////////////////////
#endif
