// -*- Mode:C++ -*-
//Header:
//File: ValVector.cc
//Author: NODA, Itsuki
//Date: 2000/01/02
//

//ModifyHistory:
// 2000/01/02: Start to create this file
//EndModifyHistory:

#include "ValVector.h"

namespace Nnn {

   //======================================================================
   // Format Parameter (used in describe)

   UInt ValVector::formatWidth = NNN_VAL_FORMAT_W ;
   UInt ValVector::formatPrecision = NNN_VAL_FORMAT_P ;

   //======================================================================
   // Management

   //----------------------------------------------------------------------
   // copy from

   ValVector & ValVector::copyFrom(const ValVector & src) {
      C_Iterator sv = src.begin() ;
      for(Iterator dv = begin() ; sv < src.end() && dv < end() ; sv++, dv++) 
	 *dv = *sv ;
      return *this ;
   }

   //----------------------------------------------------------------------
   // duplicate

   ValVector & ValVector::duplicate(Bool copyvalp) const {
      ValVector * v = new ValVector(size()) ;
      if(copyvalp) copyTo(*v) ;
      return *v ;
   } 

   //----------------------------------------------------------------------
   // randomize

   void ValVector::randomize(const Val & minval, const Val & maxval) {
      for(Iterator i = begin() ; i < end() ; i++) 
	 *i = fltRand(minval,maxval) ;
   } 

   //----------------------------------------------------------------------
   // set

   ValVector & ValVector::set(const Val & val) {
      for(Iterator i = begin() ; i < end() ; i++) 
	 *i = val ;

      return *this ;
   } 

   ValVector & ValVector::set(const Val & val, UInt from, UInt to) {
      if(from >= size()) { 
	 ITK_WRN("from index of ValVector is out of range : from=" << from
		 << " for " << *this) ;
	 from = size() ;
      } 
      if(to > size()) {
	 ITK_WRN("to index of ValVector is out of range : to=" << to
		 << " for " << *this) ;
      }
	 
      for(UInt i = from ; i < to ; i++) 
	 v(i) = val ;

      return *this ;
   } ;

   //----------------------------------------------------------------------
   // inc

   ValVector & ValVector::inc(const ValVector & v) {
      Iterator dst = begin() ;
      C_Iterator src = v.begin() ;
      for(; dst < end() ; dst++, src++)
	 *dst += *src ;

      return *this ;
   }

   //----------------------------------------------------------------------
   // mul

   ValVector & ValVector::mul(const ValVector & v) {
      Iterator dst = begin() ;
      C_Iterator src = v.begin() ;
      for(; dst < end() ; dst++, src++)
	 *dst *= *src ;

      return *this ;
   }
      
   //----------------------------------------------------------------------
   // bmask (binary mask)

   ValVector & ValVector::bmask(const ValVector & v) {
      Iterator dst = begin() ;
      C_Iterator src = v.begin() ;
      for(; dst < end() ; dst++, src++) {
	 if(*src == 0.0) *dst = 0.0 ;
      }

      return *this ;
   }
      
   //======================================================================
   // statistics

   //----------------------------------------------------------------------
   // min & max

   Val ValVector::min() const {
      Val min, max ;
      getMinMax(min,max) ;
      return min ;
   } 

   Val ValVector::max() const {
      Val min, max ;
      getMinMax(min,max) ;
      return max ;
   } 

   void ValVector::getMinMax(Val & min, Val & max) const {
      C_Iterator v = begin() ;
      min = *v ;
      max = *v ;
      for(v++ ; v < end() ; v++) {
	 if(*v < min) min = *v ;
	 if(*v > max) max = *v ;
      }
   } 

   //----------------------------------------------------------------------
   // sum & sqsum

   Val ValVector::sum() const {
      Val sum = 0.0 ;
      for(C_Iterator v = begin() ; v < end() ; v++) 
	 sum += *v ;
      return sum ;
   }

   Val ValVector::sqsum() const {
      Val sqsum = 0.0 ;
      for(C_Iterator v = begin() ; v < end() ; v++) 
	 sqsum += sq(*v) ;
      return sqsum ;
   }

   //----------------------------------------------------------------------
   // ave & div

   Val ValVector::ave() const {
      return sum() / (Val)size() ;
   } 

   void ValVector::getAveDiv(Val & ave, Val & div) const {
      Val sum = 0.0 ;
      Val sqsum = 0.0 ;

      for(C_Iterator v = begin() ; v < end() ; v++) {
	 sum += *v ;
	 sqsum += sq(*v) ;
      }

      ave = sum/(Val)size() ;
      div = (sqsum/(Val)size()) - sq(ave) ;
   }

   Val ValVector::div() const {
      Val ave, div ;
      getAveDiv(ave,div) ;
      return div ;
   } 

   Val ValVector::sdiv() const {
      return sqrt(div()) ;
   } 

   //======================================================================
   // I/O

   //----------------------------------------------------------------------
   // save

   void ValVector::save(ostream & ostr) const {
      StrForm::writeBeginTerm(ostr,LblTermHead) ;

      StrForm::writeAssocRole(ostr,LblSize) ; 
      ostr << size() ;
      StrForm::writeBeginList(ostr) ;
      for(C_Iterator i = begin() ; i < end() ; i++) {
	 if(i != begin()) StrForm::writeArgSep(ostr) ;
	 ostr << NNN_VAL_FORMAT << *i ;
      }
      StrForm::writeEndTerm(ostr) ;
   }

   void ValVector::save(ostream& ostr, 
			const SubString & indent) const {
      SubString indent2 = indent + "  " ;

      StrForm::writeBeginTerm(ostr,LblTermHead) ;

      StrForm::writeAssocRole(ostr,LblSize) ; 
      ostr << size() ;
      StrForm::writeBeginList(ostr) ;

      UInt c = 0 ;
      for(C_Iterator i = begin() ; i < end() ; i++, c++) {
	 if(i != begin()) StrForm::writeArgSep(ostr) ;
	 if(c%8 == 0) ostr << endl << indent2 ;
	 ostr << NNN_VAL_FORMAT << *i ;
      }
      StrForm::writeEndTerm(ostr) ;
   }

   void ValVector::saveAsMatrix(ostream& ostr, 
				UInt m,
				const SubString & indent) const {
      SubString indent2 = indent + "  " ;

      StrForm::writeBeginTerm(ostr,LblTermHead) ;

      StrForm::writeAssocRole(ostr,LblSize) ; 
      ostr << size() ;
      StrForm::writeBeginList(ostr) ;

      UInt c = 0 ;
      for(C_Iterator i = begin() ; i < end() ; i++, c++) {
	 c = c%m ;
	 if(i != begin()) StrForm::writeArgSep(ostr) ;
	 if(i != begin() && c == 0) ostr << endl ;
	 if(c%8 == 0) ostr << endl << indent2 ;
	 ostr << NNN_VAL_FORMAT << *i ;
      }
      StrForm::writeEndTerm(ostr) ;
   }
      
			
   
   //----------------------------------------------------------------------
   // scan

   void ValVector::scan(StrForm & sform) {
      if(!sform.isTerm(LblTermHead)) {
	 ITK_ERR("Illegal ValVector StrForm:" << sform) ;
	 throw 0 ;
      } 
      
      Index n = 0;
      sform.assocCvtUInt(LblSize,n) ;
      resize(n) ;

      Index l = sform.lengthListArg() ;
      if(l != n) ITK_WRN("mismatch ValVector size" << sform) ;

      Iterator idx = begin() ;
      for(Index i = 0 ; i < l && idx < end() ; i++, idx++) {
	 *idx = sform.nthListArg(i)->cvtFlt() ;
      }
   } ;
} ;
