// -*- Mode:C++ -*-
//Header:
//File: BasePlayer.h
//Title: Base Player for Soccer Server 7
//Author: NODA, Itsuki
//Date: 2001/03/02
//

//ModifyHistory:
// 2001/03/02: Start to create this file
//EndModifyHistory:

#ifndef _saloo_BasePlayer_h_
#define _saloo_BasePlayer_h_
////////////////////////////////////////////////////////////////////////

#include "itk/itk.h"
#include "sctypes.h"
#include "CycleStep.h"
#include "SensorInfo.h"

namespace Saloo {

   using namespace Itk ;
   using namespace Soccer ;

   //----------------------------------------------------------------------
   // constants

   //----------------------------------------
   // for debug
   
   const VbsLevel Vbs_Comm = 10 ;
   const VbsLevel Vbs_Test = 1 ;

   //======================================================================
   // Base Player

   class BasePlayer : public WithDescriber {
      public:
	 virtual const char * _classname() const { return "BasePlayer" ; } ;

	 //==================================================
	 // PlayMode
	 
	 typedef ScWord * PlayMode ;

	 //==================================================
	 // body (BasePlayer)

	 //--------------------------------------------------
	 // member
      public:

	 //------------------------------
	 // player's info

	 SString teamname ;
	 UInt unum ;
	 Bool goaliep ;
	 Side side ;

	 //------------------------------
	 // connection

	 ClientSocket socket ;

	 SString sshost ;
	 UInt ssport ;
	 Flt ssproto ;
                                           
	 Buffer sendbuf ;

	 Buffer recvbuf ;
	 SimpleSexp * recvsexp ;
	 SimpleSexp::Heap sxheap ;

	 CycleStep sendCount ;
	 CycleStep recvCount ;
	 CycleStep bodyComCount ;

	 //------------------------------
	 // Sensor Info

	 SensorInfo sinfo ;

	 //------------------------------
	 // message table

	 PlayMode _playmode ;

	 ITK_DEF_ACCESS(PlayMode & playmode(), { return _playmode ; }) ;

	 //--------------------------------------------------
	 // constructor
      public:
	 BasePlayer() {} ;

	 //--------------------------------------------------
	 // initialize
      public:
	 Bool init(const char * const shost = DfltSSHost,
		   UInt sport = DfltSSPort,
		   const char * const tname = DfltTeamName,
		   Bool gp = False,
		   Flt proto = DfltSSProto,
		   Bool openp = True,
		   Bool reconnectp = False) {
	    sshost = shost ;
	    ssport = sport ;
	    teamname = tname ;
	    goaliep = gp ;
	    ssproto = proto ;
	    side = NoSide ;
	    unum = 0 ;

	    sendbuf.alloc(DfltBufSize) ;
	    recvbuf.alloc(DfltBufSize) ;

	    sinfo.setTeamname(teamname) ;
	    
	    Bool r = False ;
	    if(openp) r = open(reconnectp) ;
	    return r ;
	 } ;

	 //--------------------------------------------------
	 // open to server
      public:
	 Bool open(Bool reconnectp = False) {
	    socket.init((const char *)sshost,Socket::UDP,ssport) ;
	    socket.setBlocking() ;
	    
	    Bool r ;
	    
	    if(!reconnectp) {
	       r = sendInit() ;
	    } else {
	       r = sendReconnect() ;
	    }

	    return r ;
	 } ;

	 //--------------------------------------------------
	 // send (init...)
      public:
	 virtual Bool sendInit() ;

	 //--------------------------------------------------
	 // send (reconnect...)
      public:
	 virtual Bool sendReconnect() ;

	 //--------------------------------------------------
	 // send commands
      public:
	 virtual Bool send() ;

	 virtual Bool sendMove(Flt x, Flt y) ;
	 Bool sendMove(const Vec & pos) { return sendMove(pos.x,pos.y) ; } ;

	 virtual Bool sendTurn(Flt power) ;
	 virtual Bool sendDash(Flt power) ;
	 virtual Bool sendKick(Flt dir, Flt power) ;
	 virtual Bool sendCatch(Flt dir) ;
	 virtual Bool sendTurnNeck(Flt pow) ;
	 virtual Bool sendSay(const SubString & msg) ;
	 virtual Bool sendSenseBody() ;
	 virtual Bool sendChangeView(ViewWidth vw, ViewQuality vq) ;

	 virtual Bool sendCommandPre(Bool bodyp) ;
	 virtual Bool sendCommandPost(Bool bodyp) ;

	 //--------------------------------------------------
	 // receive and scan sensor infomation
      public:
	 virtual Bool recv() ;

	 //--------------------------------------------------
	 // sensor cycle
      public:
	 virtual Bool cycleSensor() ;

	 //--------------------
	 // cycleSensorPre
	 /** 
	    Executed before the body part of cycleSensor().
	    If it return False, cycleSensor immediately returns.
	  */
	 virtual Bool cycleSensorPre() ;  // if return False, 

	 //--------------------
	 // cycleSensorPost
	 /** 
	    Executed after the body part of cycleSensor().
	    @arg result is return value of the body part.
	    The return value will be the return value of cycleSensor().
	  */
	 virtual Bool cycleSensorPost(Bool result) ;

	 //--------------------
	 // scanSensor

	 virtual Bool scanSensor(SimpleSexp & s) ;
	 Bool scanSensor(SimpleSexp * s) {
	    return scanSensor(*s) ;
	 } ;

	 virtual Bool scanSensorPre(SimpleSexp & s) ;
	 virtual Bool scanSensorPost(SimpleSexp & s, Bool r) ;

	 //--------------------------------------------------
	 // terminate player
      public:
	 virtual Bool terminate() ;

	 //--------------------------------------------------
	 // describe
      public:
	 virtual void describe(ostream & ostr, 
			       const Bool detailp = True) const {
	    ostr << "#" << _classname() << "[" 
		 << teamname << ":" << unum ;
	    if(goaliep) ostr << "/goalie" ;
	    ostr << "]" ;

	    if(detailp) {
	       ostr << endl ;
	       ostr << "\t" "sshost:port=" << sshost << ":" << ssport << endl ;
	       ostr << "\t" "protocol_version=" << ssproto << endl ;
	       ostr << "\t" "side=" << sideString(side) << endl ;
	       describeAux(ostr,detailp) ;
	    } 
	 } ;

	 virtual void describeAux(ostream & ostr, 
				  const Bool detailp = True) const {
	    return ;
	 } ;
   } ;

} ;

////////////////////////////////////////////////////////////////////////
#endif

